"""
Module containing the File object for files in a gist
"""

from typing import Optional, TypeVar, Dict, List


__all__ = ("File",)

F = TypeVar("F", bound="File")


class File:
    """The file object that is provided when editing and creating gists"""

    def __init__(
        self,
        *,
        name: str,
        content: Optional[str] = None,
        new_name: Optional[str] = None
    ):
        self.name: str = name
        self.content: str = content

        self.new_name: str = new_name or self.name

    def __eq__(self, other) -> bool:
        return (
            isinstance(other, File)
            and self.name == other.name
            and self.content == other.content
        )

    def to_dict(self) -> Dict:
        """Returns the dictionary form of the File object"""

        files_dict = {self.name: {"filename": self.new_name}}

        if self.content:
            content_dict = {"content": self.content}
            files_dict[self.name].update(content_dict)

        return files_dict

    @classmethod
    def from_dict(cls, files_dict: Dict) -> List[F]:
        """Returns a list of File objects from a files dictionary"""

        # Example structure of a files_dict:
        #     {
        # "file1.txt": {
        #   "content": ""
        # },
        # "file2.txt": {
        #   "filename": "file2.txt",
        #   "content": "test"
        # },
        # }

        file_objs = []
        for name in files_dict.keys():
            self = cls.__new__(cls)

            self.name = name
            self.content = files_dict[name].get("content", None)
            self.new_name = files_dict[name].get("filename", self.name)

            file_objs.append(self)

        return list(sorted(file_objs, key=lambda f: f.name))



"""
Module containing the main Client class used to send requests
"""


from typing import Optional, Dict, List
import sys

import aiohttp

from .gist import Gist
from .file import File
from .exceptions import HTTPException, AuthorizationFailure, NotFound
from .constants import API_URL


__all__ = ("Client",)


class Client:
    """Does not take access token directly to allow actions that do not require authorization.

    Use the authorize method to authorize
    """

    def __init__(self):
        self.access_token = None  # Set by Client.authorize()

        self.user_data = None

    async def authorize(self, access_token: str):
        """Method used to authorize the client,
        In order to send requests that need authorization
        """

        self.access_token = access_token

        # TODO User object rather than the raw json data
        self.user_data = await self.fetch_user_data()

    async def request(
        self,
        method: str,
        url: str,
        *,
        params=None,
        data=None,
        headers=None,
        authorization: bool = True,
    ) -> Dict:
        """The method to make asynchronous requests to the GitHub API"""

        headers_final = {
            "Accept": "application/vnd.github.v3+json",
            "User-Agent": (
                f"Gists.py (https://github.com/witherredaway/gists.py) "
                f"Python/{sys.version_info[0]}.{sys.version_info[1]} aiohttp/{aiohttp.__version__}"
            ),
        }
        if authorization:
            if not self.access_token:
                raise AuthorizationFailure(
                    "To use functions that require authorization, "
                    "please authorize the Client with Client.authorize"
                )
            headers_final["Authorization"] = f"token {self.access_token}"

        request_url = f"{API_URL}/{url}"

        if headers is not None and isinstance(headers, dict):
            headers_final.update(headers)

        async with aiohttp.ClientSession() as session:
            response = await session.request(
                method, request_url, params=params, json=data, headers=headers_final
            )

            try:
                data = await response.json()
            except aiohttp.client_exceptions.ContentTypeError:
                data = response.content

            remaining = response.headers.get("X-Ratelimit-Remaining")

            if 300 > response.status >= 200:
                return data
            elif response.status == 429 or remaining == "0":
                raise HTTPException(response, data)
            elif response.status == 404:
                raise NotFound(response, data)
            elif response.status == 401:
                raise AuthorizationFailure(
                    "Invalid personal access token has been passed."
                )

    async def fetch_user_data(self) -> Dict:
        """Fetch data of the authenticated user"""

        try:
            # TODO User object rather than the raw json data
            user_data: Dict = await self.request("GET", "user", authorization=True)
        except NotFound as error:
            raise NotFound(error.response, "User not found")

        return user_data

    async def fetch_gist_data(self, gist_id: str) -> Dict:
        """Fetch data of a Gist"""

        try:
            gist_data: Dict = await self.request(
                "GET", f"gists/{gist_id}", authorization=False
            )
        except NotFound as error:
            raise NotFound(error.response, "Gist not found")
        return gist_data

    async def get_gist(self, gist_url_or_id: str) -> Gist:
        """Get a Gist object representing the gist associated with the provided gist ID or url

        Does not require authorization.
        """

        gist_id = Gist.gist_url_to_id(gist_url_or_id)

        data = await self.fetch_gist_data(gist_id)
        return Gist(data, self)

    async def create_gist(
        self,
        *,
        files: List[File],
        description: str = None,
        public: bool = True,
    ) -> Gist:
        """Create a new gist and return a Gist object associated with that gist"""

        files_dict = {}
        for file in files:
            # Update the files_dict with the dictionaries of each File object
            files_dict.update(file.to_dict())

        data = {"public": public, "files": files_dict}
        params = {"scope": "gist"}

        if description:
            data["description"] = description

        gist_data = await self.request("POST", "gists", data=data, params=params)
        return Gist(gist_data, self)

    async def update_gist(self, gist_url_or_id: str):
        """Alias of fetch_gist_data, used to fetch a gist's data."""

        gist_id = Gist.gist_url_to_id(gist_url_or_id)

        updated_gist_data = await self.fetch_gist_data(gist_id)
        return updated_gist_data

    async def edit_gist(
        self,
        gist_url_or_id: str,
        *,
        description: Optional[str] = None,
        files: Optional[List[File]] = None,
    ) -> Dict:
        """Edit the gist associated with the provided gist id, and return the edited data"""

        gist_id = Gist.gist_url_to_id(gist_url_or_id)

        data = {}

        if files:
            files_dict = {}
            for file in files:
                # Update the files_dict with the dictionaries of each File object
                files_dict.update(file.to_dict())
            data["files"] = files_dict

        if description:
            data["description"] = description

        try:
            edited_gist_data = await self.request(
                "PATCH", f"gists/{gist_id}", data=data
            )
        except NotFound as error:
            raise NotFound(error.response, "Gist not found")
        return edited_gist_data

    async def delete_gist(self, gist_url_or_id: str):
        """Delete the gist associated with the provided gist id"""

        gist_id = Gist.gist_url_to_id(gist_url_or_id)

        try:
            await self.request("DELETE", f"gists/{gist_id}")
        except NotFound as error:
            raise NotFound(error.response, "Gist not found")



"""
Module containing the exceptions that the wrapper uses
"""


from typing import Optional, Union, Dict, Any

import aiohttp

__all__ = (
    "ClientException",
    "HTTPException",
    "AuthorizationFailure",
    "NotFound",
)


class GistsException(Exception):
    """Base exception class"""

    pass


class ClientException(GistsException):
    """Exception that is raised when an operation in the Client class fails"""

    pass


class GistException(GistsException):
    """Exception that is raised when an operation in the Client class fails"""

    pass


class HTTPException(ClientException):
    """Exception that is raised for HTTP request related failures"""

    def __init__(
        self,
        response: aiohttp.ClientResponse,
        data: Optional[Union[str, Dict[str, Any]]],
    ):
        self.response: aiohttp.ClientResponse = response
        self.status: int = response.status
        self.code: int
        self.text: str

        if isinstance(data, dict):
            self.code = data.get("code", 0)

            self.text = data.get("message", "")
        else:
            self.text = data or ""
            self.code = 0

        message = f"{self.response.status} {self.response.reason} ({self.code})"
        if len(self.text):
            message += f": {self.text}"

        super().__init__(message)


class AuthorizationFailure(ClientException):
    """Exception that is raised when authorizing provided token in the Client class fails."""

    pass


class NotFound(HTTPException):
    """HTTPException that is raised when the status code is 404"""

    pass


class DataFetchError(ClientException):
    """Exception that is raised when fetching data fails."""

    pass



"""
Module containing constants used throughout the wrapper
"""

import re


API_URL = "https://api.github.com"

TIME_FORMAT = "%Y-%m-%dT%H:%M:%SZ %z"

GIST_URL_REGEX = re.compile(
    "(?:https:\/\/)?gist\.github\.com\/(?P<username>.+)\/(?P<gist_id>[\d\w]+)|(?:https:\/\/)?gist\.github\.com\/(?P<gist_id2>[\d\w]+)"
)



"""
Module containing the Gist object
"""
from __future__ import annotations

import typing
from typing import Optional, Dict, List
import datetime

from .file import File
from .constants import TIME_FORMAT, GIST_URL_REGEX

if typing.TYPE_CHECKING:
    from .client import Client


__all__ = ("Gist",)


class Gist:
    """The Gist object that represents a gist"""

    __slots__ = (
        "client",
        "comments",
        "comments_url",
        "commits_url",
        "_created_at",
        "_description",
        "description",
        "_files",
        "files",
        "forks",
        "forks_url",
        "git_pull_url",
        "git_push_url",
        "history",
        "url",
        "id",
        "node_id",
        "owner",
        "public",
        "truncated",
        "_updated_at",
        "api_url",
        "user",
    )

    def __init__(self, data: Dict, client: Client):
        self.client = client

        self._update_attrs(data)

    def __eq__(self, other):
        """Equality comparison. Compares:
        - Types of the two objects
        - Description of the two Gist objects
        - Files of the two Gist objects
            - Files' names
            - Files' contents
        """
        if not (
            isinstance(other, Gist)
            and self.description == other.description
            and len(self.files) == len(other.files)
        ):
            return False

        for self_file, other_file in zip(self.files, other.files):
            if self_file != other_file:
                return False
        return True

    def _update_attrs(self, data: Dict):
        """Update the Gist object's attributes with the provided data"""

        self.comments: int = data.get("comments", None)
        self.comments_url: str = data.get("comments_url", None)
        self.commits_url: str = data.get("commits_url", None)
        self._created_at: str = data.get("created_at", None)
        self._description: str = data.get("description", None)
        self._files: Dict = data.get("files", None)
        self.forks: List = data.get("forks", None)  # TODO Fork object
        self.forks_url: str = data.get("forks_url", None)
        self.git_pull_url: str = data.get("git_pull_url", None)
        self.git_push_url: str = data.get("git_push_url", None)
        self.history: List = data.get("history", None)  # TODO History object
        self.url: str = data.get("html_url", None)
        self.id: str = data.get("id", None)
        self.node_id: str = data.get("node_id", None)
        self.owner: Dict = data.get("owner", None)  # TODO User object
        self.public: bool = data.get("public", None)
        self.truncated: bool = data.get("truncated", None)
        self._updated_at: str = data.get("updated_at", None)
        self.api_url: str = data.get("url", None)
        self.user: None = data.get("user", None)

        self.description: str = self._description
        self.files: List[File] = File.from_dict(self._files)

    @staticmethod
    def _get_dt_obj(time: str) -> datetime.datetime:
        """Internal method to convert string datetime format to datetime object"""
        time = time + " +0000"  # Tells datetime that the timezone is UTC
        dt_obj: datetime.datetime = datetime.datetime.strptime(time, TIME_FORMAT)
        return dt_obj

    @property
    def created_at(self) -> datetime.datetime:
        return self._get_dt_obj(self._created_at)

    @created_at.setter
    def created_at(self, value: str):
        self._created_at = value

    @property
    def updated_at(self) -> datetime.datetime:
        return self._get_dt_obj(self._updated_at)

    @updated_at.setter
    def updated_at(self, value: str):
        self._updated_at = value

    async def update(self):
        """Fetch and update the Gist object with the gist"""

        updated_gist_data = await self.client.update_gist(self.id)
        self._update_attrs(updated_gist_data)

    async def edit(
        self, *, description: Optional[str] = None, files: Optional[List[File]] = None
    ):
        """Edit the gist associated with the Gist object, then update the Gist object"""

        # Since both files and description are optional arguments,
        # you can edit a file through Gist.files (e.g. Gist.files[0].content = "Edited content")
        # or the description through Gist.description (e.g. Gist.description = "Edited description")
        # and then call Gist.edit() without passing any args to sync the gist with the changed values.
        if not description:
            description = self.description

        if not files:
            files = self.files

        files_dict = {}
        for file in files:
            # Update the files_dict with the dictionaries of each File object
            files_dict.update(file.to_dict())
        # If there are no changes, the gist will not edit
        if all(
            (
                self.description == self._description,
                files_dict == self._files,
            )
        ):
            return

        kwargs = {"description": description, "files": files}

        edited_gist_data = await self.client.edit_gist(self.id, **kwargs)
        self._update_attrs(edited_gist_data)

    async def delete(self):
        """Delete the gist associated with the Gist object, then delete the Gist object itself"""

        await self.client.delete_gist(self.id)

    @staticmethod
    def gist_url_to_id(url_or_id: str) -> str:
        """Function to get a gist's ID from its url, or assumes that the input was the ID itself if it is not a url"""

        match = GIST_URL_REGEX.search(url_or_id)
        if match:
            url_or_id = match.group("gist_id") or match.group("gist_id2")

        return url_or_id
